import os
from remote_graphcast.gcutils.cdsutils import *

def test_saves_correctly():
	filename = "tmp.cruft"
	save_cds_file("key", "url", filename)
	with open(filename) as f:
		content = f.read()
	
	assert content == "key: key\nurl: url\n"
	os.remove(filename)


import logging

def setup_logging():
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    logging.basicConfig(
        format='%(levelname)s: %(asctime)s %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
        level=logging.INFO
    )


    


from remote_graphcast.gcutils.log_config import setup_logging
setup_logging()
from .remote_cast import cast_from_parameters
import sys

if __name__ == '__main__':
	# the first passed in argument is a filename
	cast_from_parameters(sys.argv[1])


import boto3
import time
import runpod
import json
from remote_graphcast.gcutils.constants import *
from remote_graphcast.gcutils.inpututils import *
import logging
logger = logging.getLogger(__name__)

class UploadMonitor():
	def __init__(self, pod, aws_access_key_id, aws_secret_access_key, aws_bucket, cast_id) -> None:
		self.s3_client = boto3.client('s3', aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key)
		self.pod = pod
		self.cast_id = cast_id
		self.aws_bucket = aws_bucket

	def is_complete(self):
		pod = runpod.get_pod(self.pod['id'])

		if not pod:
			raise Exception(f'runpod pod {self.pod} was terminated prematurely, output of runpod.get_pod is: {pod}')

		try:		
			self.s3_client.head_object(Bucket=self.aws_bucket, Key=get_completion_path(self.cast_id))
			return True
		except Exception as e:
			if hasattr(e, 'response') and e.response['Error'].get('Code') == "404":
				logger.debug(f'upload not complete, expected_s3_error {e}')
				return False
			else: 
				raise e

	def upload_location(self):
		return f"s3://{self.aws_bucket}/{self.cast_id}/"

def cast_from_parameters(param_file=None, **kwargs):
	if param_file is not None:
		assert param_file.endswith(".json"), "param_file must be a json file"

		with open(param_file, "r") as f:
			parameters = json.load(f)
		kwargs = parameters

	remote_cast(**kwargs)

def validate_gpu_type_id(gpu_type_id):
	if gpu_type_id == "NVIDIA A100 80GB PCIe":
		logger.warn(f'{gpu_type_id} is known to crash on runpod around 50% of the time when used in combination with the remote graphcast docker image. We suggest using NVIDIA A100-SXM4-80GB instead')

def validate(gpu_type_id, forcast_list, strict_start_times):
	validate_gpu_type_id(gpu_type_id)
	validate_forcast_list(forcast_list, strict_start_times)

def remote_cast(
		aws_access_key_id, 
		aws_secret_access_key, 
		aws_bucket, 
		cds_url, 
		cds_key, 
		forcast_list, 
		runpod_key,
		cast_id=None,
		gpu_type_id="NVIDIA A100-SXM4-80GB", # graphcast needs at least 61GB GPU ram
		container_disk_in_gb=50,
		strict_start_times=True
	):	

	if cast_id is None:
		cast_id = generate_cast_id()
		logger.info(f'cast_id generated {cast_id}')
	
	logger.info(f'validating input')
	validate(gpu_type_id, forcast_list, strict_start_times)
	runpod.api_key = runpod_key
	
	pod = runpod.create_pod(
		cloud_type="SECURE", # or else someone might snoop your session and steal your AWS/CDS credentials
		name=f"easy-graphcast-{cast_id}", 
		image_name="jonathanle67/easy-graphcast-image:latest", 
		gpu_type_id=gpu_type_id,
		container_disk_in_gb=container_disk_in_gb,
		env={
			AWS_ACCESS_KEY_ID: aws_access_key_id,
			AWS_SECRET_ACCESS_KEY: aws_secret_access_key,
			AWS_BUCKET: aws_bucket,
			CDS_KEY: cds_key,
			CDS_URL: cds_url,
			FORCAST_LIST: forcast_list,
			CAST_ID: cast_id			
		}
	)
	
	logger.info(f"forcasting pod created, id: {pod['id']}, machineId: {pod['machineId']}, machine: {pod['machine']}")
	monitor = UploadMonitor(pod, aws_access_key_id, aws_secret_access_key, aws_bucket, cast_id)

	while not monitor.is_complete():
		logger.info('checking runpod and s3 for forcast status: all systems green')
		time.sleep(60)

	logger.info(f'easy-graphcast forcast is complete saved to, {monitor.upload_location()}')

	runpod.terminate_pod(pod['id'])

	logger.info('pod terminated')

	return monitor.upload_location()





import requests
from requests.exceptions import RequestException
import os
import yaml
from datetime import datetime
import logging

logger = logging.getLogger(__name__)


RCFILE_PATH = "~/.cdsapirc"

def save_cds_rcfile(cds_key, cds_url):
	save_cds_file(cds_key, cds_url, RCFILE_PATH)

def save_cds_file(cds_key, cds_url, filename):

	expanded_filename = os.path.expanduser(filename)
	with open(expanded_filename, "w") as f:
		data = {
			'key': cds_key,
			'url': cds_url
		}
		yaml.dump(data, f)

def get_latest_available_date(api_url="https://cds.climate.copernicus.eu/api/v2.ui/resources/reanalysis-era5-single-levels", retries=3, timeout=5):
    for attempt in range(retries):
        try:
            result = requests.get(api_url, timeout=timeout)
            result.raise_for_status()

            data = result.json()

            date_range = data.get('structured_data', {}).get('temporalCoverage', '')
            if not date_range:
                raise ValueError("Temporal coverage not found in data")

            final_date = date_range.split('/')[1]

            return datetime.strptime(final_date, "%Y-%m-%d")

        except RequestException as e:
            logger.error(f"CDS most recent date request failed: {e}")
            if attempt == retries - 1:
                raise
        except ValueError as e:
            logger.error(f"cds most recent date data parsing error: {e}")
            raise

    raise Exception("Maximum retries reached")


AWS_ACCESS_KEY_ID = 'AWS_ACCESS_KEY_ID' 
AWS_SECRET_ACCESS_KEY = 'AWS_SECRET_ACCESS_KEY'
AWS_BUCKET = 'AWS_BUCKET'
CDS_URL = 'CDS_URL'
CDS_KEY = 'CDS_KEY'
FORCAST_LIST = 'GRAPHCAST_FORCAST_LIST'
CAST_ID = 'CAST_ID'
RUNPOD_KEY = "RUNPOD_KEY"


# from https://github.com/ecmwf-lab/ai-models-graphcast/blob/main/ai_models_graphcast/input.py 
CF_NAME_SFC = {
    "10u": "10m_u_component_of_wind",
    "10v": "10m_v_component_of_wind",
    "2t": "2m_temperature",
    "lsm": "land_sea_mask",
    "msl": "mean_sea_level_pressure",
    "tp": "total_precipitation_6hr",
    "z": "geopotential_at_surface",
}


import os
import logging
import json
from datetime import datetime, timedelta
import random
from remote_graphcast.gcutils.cdsutils import get_latest_available_date

logger = logging.getLogger(__name__)

def generate_cast_id():
	adj = [
		"whispering", "forgotten", "sublime", "phantom", "mystic",
		"cosmic", "dazzling", "spectral", "fleeting", "shimmering",
		"invisible", "echoing", "intangible", "melodic", "transient",
		"prismatic", "noctilucent", "aurora", "enigmatic", "ephemeral",
		"stellar", "mythic", "azure", "crystalline", "mystifying",
		"radiant", "lucent", "polar", "celestial", "timeless",
		"cobalt", "harmonic", "infinite", "nautical", "primal",
		"solar", "wandering", "auroral", "floral", "lucent",
		"maritime", "opaline", "stellar", "twinkling", "zodiacal",
		"botanic", "coral", "dreamlike", "glacial", "harmonic",
		"ethereal", "arcane", "sylvan", "emerald", "golden",
		"silver", "moonlit", "sunlit", "twilight", "midnight",
		"dawn", "dusk", "summer", "winter", "autumn",
		"spring", "ancient", "modern", "timeless", "endless",
		"boundless", "ageless", "eternal", "everlasting", "perpetual",
		"unending", "infinite", "limitless", "unbounded", "unfathomable"
	]
	nouns = [
		"cosmos", "phantasm", "stardust", "galaxy", "oasis",
		"tempest", "artifact", "cascade", "maelstrom", "saga",
		"reverie", "illusion", "odyssey", "riddle", "miracle",
		"cosmic", "haven", "mirage", "oracle", "tapestry",
		"atlantis", "harmony", "nebulae", "pinnacle", "relic",
		"zenith", "borealis", "inferno", "oceanus", "pulsar",
		"saga", "allegory", "clarity", "expanse", "fable",
		"lagoon", "narrative", "pastoral", "rhapsody", "sonnet",
		"tundra", "cascade", "echo", "infinity", "monolith",
		"oasis", "panorama", "sphere", "torrent", "wilderness",
		"abyss", "beacon", "chronicle", "dimension", "eclipse",
		"fjord", "glacier", "horizon", "island", "jungle",
		"kaleidoscope", "labyrinth", "meadow", "nirvana", "ocean",
		"paradise", "quasar", "rainforest", "sanctuary", "tide",
		"universe", "vortex", "waterfall", "xanadu", "yonder",
		"zenith", "arcadia", "bazaar", "citadel", "dome",
		"enclave", "fortress", "grove", "hamlet", "isle"
	]

	present = datetime.now()
	time_string = present.strftime("%Y-%m-%d_%H-%M-%S")

	return f"{time_string}_{random.choice(adj)}_{random.choice(nouns)}"


def confirm_start_time_exists(start_point, c):
	start_datetime = datetime.strptime(start_point['start_date'], '%Y%m%d')

	c.retrieve(
		"reanalysis-era5-single-levels",
		{
			'product_type': 'reanalysis',
			'format': 'netcdf',
			'day': start_datetime.day, 
			'year': start_datetime.year,
			'month': start_datetime.month,
			'time': start_point['start_time'],
			'variable': '2m_temperature',
			'area': [1, -1, -1, 1,],
		},
		'file.nc'
	)

	os.remove('file.nc')


def parse_forcast_list(forcast_list):
	# the runpod API cannot handle double quotes so forcast_list is single quoted
	# we need to convert it to regular JSON
	forcast_list = json.loads(forcast_list.replace("'", '"')) 

	for start in forcast_list:
		start['start_time'] = int(start['start'][-2:])
		start['start_date'] = start['start'][:-2]

	return forcast_list
	
def get_completion_path(cast_id):
	return f"{cast_id}/.easy_graphcast_complete"

def validate_forcast_list(forcast_list, strict_start_times=True):	
	# the runpod API cannot handle double quotes so forcast_list is single quoted
	if '"' in forcast_list:
		raise ValueError('forcast_list cannot contain double quotes (this is a limitation of the runpod API) replace with single quotes.')

	forcast_list = parse_forcast_list(forcast_list)

	if strict_start_times:
		for start in forcast_list:
			if start['start_time'] not in [6, 18]:
				raise ValueError('you must start all graphcast forcasts at either 0600 or 1800 (see https://youtu.be/PD1v5PCJs_o?t=1915 for more information). You can disable this check by setting strict_start_times=False')
	try:
		latest_available_date = get_latest_available_date()	
		got_date = True
	except Exception as e:
		logger.info(f"could not get the latest available date due to error: {e}")
		got_date = False

	if got_date:
		for start in forcast_list:
			start_date = datetime.strptime(start['start'], '%Y%m%d%H')
			if latest_available_date < start_date:
				raise ValueError(f"there is no public ERA5 data for the start date {start['start']} so you cannot start a forcast from that date. The latest available date is {latest_available_date}")
	else:
		latest_available_date = datetime.now() - timedelta(days=5)
		for start in forcast_list:
			start_date = datetime.strptime(start['start'], '%Y%m%d%H')
			if latest_available_date < start_date:
				logger.warning(f"the forcast start date {start_date} is more recent than 5 days ago, the data may not exist yet. Monitor the runpod instance closely")

	logger.info('input passed validation')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="remote-graphcast-runpod",
    version="0.0.20",
    author="Louka Ewington-Pitsos",
    author_email="lewingtonpitsos@gmail.com",
    description="Allows people without massive GPUs to easily run graphcast on remote runpod servers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CisgenX/remote-graphcast.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'boto3', 
        'runpod'
	],
    python_requires='>=3.10',
)